'use strict';

var path = require('path');
var findRoot = require('find-root');
var chalk = require('chalk');
var _ = require('lodash');

function DuplicatePackageCheckerPlugin() {}

function cleanPath(path) {
  return '.' + path.split('/node_modules/').join('/~/');
}

// Get closest package definition from path
function getClosestPackage(modulePath) {
  var root = void 0;
  var pkg = void 0;

  // Catch findRoot or require errors
  try {
    root = findRoot(modulePath);
    pkg = require(path.join(root, 'package.json'));
  } catch (e) {
    return null;
  }

  // If the package.json does not have a name property, try again from
  // one level higher.
  // https://github.com/jsdnxx/find-root/issues/2
  // https://github.com/date-fns/date-fns/issues/264#issuecomment-265128399
  if (!pkg.name) {
    return getClosestPackage(path.resolve(root, '..'));
  }

  return {
    package: pkg,
    path: root
  };
}

DuplicatePackageCheckerPlugin.prototype.apply = function (compiler) {
  compiler.plugin('emit', function (compilation, callback) {

    var context = compilation.compiler.context;
    var modules = {};

    compilation.modules.forEach(function (module) {

      if (!module.resource) {
        return;
      }

      var pkg = void 0;
      var packagePath = void 0;

      var closestPackage = getClosestPackage(module.resource);

      // Skip module if no closest package is found
      if (!closestPackage) {
        return;
      }

      pkg = closestPackage.package;
      packagePath = closestPackage.path;

      var modulePath = cleanPath(packagePath.replace(context, ''));
      var version = pkg.version;

      modules[pkg.name] = modules[pkg.name] || [];

      var isSeen = _.find(modules[pkg.name], function (module) {
        return module.version === version;
      });

      if (!isSeen) {
        modules[pkg.name].push({ version: version, path: modulePath });
      }
    });

    var duplicates = _.omitBy(modules, function (versions) {
      return versions.length <= 1;
    });

    if (Object.keys(duplicates).length) {

      var error = 'duplicate-package-checker:';

      _.each(duplicates, function (instances, name) {
        instances = instances.map(function (version) {
          return chalk.green.bold(version.version) + ' ' + chalk.white.bold(version.path);
        });
        error += '\n  ' + chalk.yellow.bold('<') + chalk.green.bold(name) + chalk.yellow.bold('>') + '\n';
        error += '    ' + instances.join('\n    ') + '\n';
      });

      compilation.warnings.push(new Error(error));
    }

    callback();
  });
};

module.exports = DuplicatePackageCheckerPlugin;